/**
* @file Estrangela ASCII code font utilities
* @version 1.1.0
* @author Greg Borota
* @copyright (c) 2017 Greg Borota.
* @license MIT
*
* Permission is hereby granted, free of charge, to any person obtaining a copy
* of this software and associated documentation files (the "Software"), to deal
* in the Software without restriction, including without limitation the rights
* to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
* copies of the Software, and to permit persons to whom the Software is
* furnished to do so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in
* all copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
*/

// https://peshitta.github.io
// http://www.peshitta.org/initial/standard.html
// http://www.peshitta.org/initial/software.html
// http://cal1.cn.huc.edu/searching/fullbrowser.html

import { clearDotting, getSort, hasDotting } from 'aramaic-mapper';

var obj;
/** @module estrangelaCodeUtil */
/**
 * Estrangela consonant name to value map
 * @constant
 * @type { Object.<string, string> }
*/
var consonantsByName = Object.freeze(
  Object.create(null, {
    alaph: { value: '0', enumerable: true },
    beth: { value: 'b', enumerable: true },
    gamal: { value: 'g', enumerable: true },
    dalath: { value: 'd', enumerable: true },

    he: { value: 'h', enumerable: true },
    waw: { value: 'w', enumerable: true },
    zayn: { value: 'z', enumerable: true },

    heth: { value: 'x', enumerable: true },
    teth: { value: '=', enumerable: true },
    yod: { value: 'y', enumerable: true },

    kaph: { value: 'k', enumerable: true },
    lamadh: { value: 'l', enumerable: true },
    mim: { value: 'm', enumerable: true },
    nun: { value: 'n', enumerable: true },

    semkath: { value: 's', enumerable: true },
    e: { value: '9', enumerable: true },
    pe: { value: 'p', enumerable: true },
    sadhe: { value: 'c', enumerable: true },

    qoph: { value: 'q', enumerable: true },
    resh: { value: 'r', enumerable: true },
    shin: { value: '4', enumerable: true },
    taw: { value: 't', enumerable: true }
  })
);

var l = consonantsByName;
/**
 * Estrangela base consonants - the 22 aramaic consonants
 * @constant
 * @type { string[] }
*/
var baseConsonants = Object.freeze([
  // abgad
  l.alaph,
  l.beth,
  l.gamal,
  l.dalath,

  // hawaz
  l.he,
  l.waw,
  l.zayn,

  // ḥaṭy
  l.heth,
  l.teth,
  l.yod,

  // kalman
  l.kaph,
  l.lamadh,
  l.mim,
  l.nun,

  // saʿpac
  l.semkath,
  l.e,
  l.pe,
  l.sadhe,

  // qarshat
  l.qoph,
  l.resh,
  l.shin,
  l.taw
]);

/**
 * Estrangela final connected modified consonants
 * @constant
 * @type { string[] }
*/
var finalConnected = Object.freeze([
  'B',
  'G',

  'X',
  '+',
  'Y',

  'K',
  'L',
  'M',
  'N',

  'S',
  '(',
  'P',

  'Q',
  '$'
]);

/**
 * Estrangela final non-connected modified consonants
 * @constant
 * @type { string[] }
*/
var finalNonConnected = Object.freeze(['<', 'J']);

/**
 * Estrangela 2 ligatures
 * @constant
 * @type { string[] }
*/
var ligatures = Object.freeze(['f', 'F']);

/**
 * All Estrangela font consonants
 * @constant
 * @type { string[] }
 */
var allConsonants = Object.freeze(
  baseConsonants
    .concat(finalConnected)
    .concat(finalNonConnected)
    .concat(ligatures)
);

/**
 * Estrangela consonants unconnected on the left
 * @constant
 * @type { string[] }
 */
var unconnected = Object.freeze([
  '0',
  'd',
  'h',
  'w',
  'z',
  'c',
  'r',
  't'
]);

/**
 * Estrangela vowel name to value map
 * @constant
 * @type { Object.<string, string> }
*/
var vowelsByName = Object.freeze(
  Object.create(null, {
    pthaha: { value: 'e', enumerable: true },
    zqapha: { value: 'a', enumerable: true },
    rbasa: { value: 'o', enumerable: true },
    hbasaEsasa: { value: ';', enumerable: true },

    zlama: { value: 'i', enumerable: true },
    rwaha: { value: 'u', enumerable: true }
  })
);

var v = vowelsByName;
/**
 * Estrangela common vowels - common to both eastern and western
 * @constant
 * @type { string[] }
*/
var commonVowels = Object.freeze([
  v.pthaha, // a
  v.zqapha, // o
  v.rbasa, // e
  v.hbasaEsasa // i u
]);

/**
 * Estrangela eastern only vowels
 * @constant
 * @type { string[] }
*/
var easternVowels = Object.freeze([
  v.zlama, // E
  v.rwaha // O
]);

/**
 * Estrangela shifted vowels (SHIFT key on)
 * @constant
 * @type { string[] }
*/
var shiftedVowels = Object.freeze([
  'E', // a
  'A', // o
  'O', // e
  'I' // E
]);

/**
 * Estrangela vowels (both common and eastern) - excluding shifted
 * @constant
 * @type { string[] }
*/
var baseVowels = Object.freeze(commonVowels.concat(easternVowels));

/**
 * All Estrangela vowels (including shifted)
 * @constant
 * @type { string[] }
 */
var allVowels = Object.freeze(baseVowels.concat(shiftedVowels));

/**
 * Sedra/CAL diacritic name map
 * 1. qushaya: __2__ - dot above
 * 2. rukkakha: __3__ - dot below
 * 3. lineaOccultans: **5** - linea occultans
 * 4. seyame: __*__ - seyame, rebwe
 * @constant
 * @type { Object.<string, string> }
*/
var diacriticsByName = Object.freeze(
  Object.create(null, {
    qushaya: { value: '2', enumerable: true },
    rukkakha: { value: '3', enumerable: true },
    lineaOccultans: { value: '5', enumerable: true },
    seyame: { value: '6', enumerable: true }
  })
);

var d = diacriticsByName;
/**
 * Estrangela base/common diacritics (both western and eastern)
 * @constant
 * @type { string[] }
*/
var commonDiacritics = Object.freeze([
  d.qushaya, // sedra/cal ' - Qushaya
  d.rukkakha, // sedra/cal , - Rukkakha
  d.lineaOccultans, // sedra/cal _ - Line under
  d.seyame // sedra/cal * - Seyame
]);

/**
 * Estrangela eastern specific diacritics
 * @constant
 * @type { string[] }
*/
var easternDiacritics = Object.freeze(['`', '1', '8']);

/**
 * Estrangela diacritics - excluding shifted
 * @constant
 * @type { string[] }
*/
var baseDiacritics = Object.freeze(
  commonDiacritics.concat(easternDiacritics)
);

/**
 * Estrangela shifted diacritics
 * @constant
 * @type { string[] }
*/
var shiftedDiacritics = Object.freeze([
  '#', // sedra/cal , - Rukkakha
  '%', // sedra/cal _ - Line under
  '^' // sedra/cal * - Seyame
]);

/**
 * All Estrangela diacritics (including shifted)
 * @constant
 * @type { string[] }
 */
var allDiacritics = Object.freeze(
  baseDiacritics.concat(shiftedDiacritics)
);

/**
 * Estrangela punctuation
 * @constant
 * @type { string[] }
*/
var punctuation = Object.freeze(['.', '?']);

/**
 * Estrangela joiners
 * @constant
 * @type { string[] }
*/
var joiners = Object.freeze(['-', '_']);

/**
 * Estrangela to ordinal ASCII value. Used for sorting:
 * a b c d e f g h i j k l m n o p q r s t u v
 * w x y z { | }
 * @constant
 * @type { Object.<string, string> }
*/
var letterAsciiMap = Object.freeze(
  Object.create(null, ( obj = {}, obj[l.alaph] = { value: 'a', enumerable: true }, obj[l.beth] = { value: 'b', enumerable: true }, obj.B = { value: 'b', enumerable: true }, obj[l.gamal] = { value: 'c', enumerable: true }, obj.G = { value: 'c', enumerable: true }, obj[l.dalath] = { value: 'd', enumerable: true }, obj[l.he] = { value: 'e', enumerable: true }, obj[l.waw] = { value: 'f', enumerable: true }, obj[l.zayn] = { value: 'g', enumerable: true }, obj[l.heth] = { value: 'h', enumerable: true }, obj.X = { value: 'h', enumerable: true }, obj[l.teth] = { value: 'i', enumerable: true }, obj['+'] = { value: 'i', enumerable: true }, obj[l.yod] = { value: 'j', enumerable: true }, obj.Y = { value: 'j', enumerable: true }, obj[l.kaph] = { value: 'k', enumerable: true }, obj.K = { value: 'k', enumerable: true }, obj['<'] = { value: 'k', enumerable: true }, obj[l.lamadh] = { value: 'l', enumerable: true }, obj.L = { value: 'l', enumerable: true }, obj[l.mim] = { value: 'm', enumerable: true }, obj.M = { value: 'm', enumerable: true }, obj[l.nun] = { value: 'n', enumerable: true }, obj.N = { value: 'n', enumerable: true }, obj.J = { value: 'n', enumerable: true }, obj[l.semkath] = { value: 'o', enumerable: true }, obj.S = { value: 'o', enumerable: true }, obj[l.e] = { value: 'p', enumerable: true }, obj['('] = { value: 'p', enumerable: true }, obj[l.pe] = { value: 'q', enumerable: true }, obj.P = { value: 'q', enumerable: true }, obj[l.sadhe] = { value: 'r', enumerable: true }, obj[l.qoph] = { value: 's', enumerable: true }, obj.Q = { value: 's', enumerable: true }, obj[l.resh] = { value: 't', enumerable: true }, obj[l.shin] = { value: 'u', enumerable: true }, obj.$ = { value: 'u', enumerable: true }, obj[l.taw] = { value: 'v', enumerable: true }, obj.f = { value: 'la', enumerable: true }, obj.F = { value: 'ta', enumerable: true }, obj[v.pthaha] = { value: 'w', enumerable: true }, obj.E = { value: 'w', enumerable: true }, obj[v.zqapha] = { value: 'x', enumerable: true }, obj.A = { value: 'x', enumerable: true }, obj[v.rbasa] = { value: 'y', enumerable: true }, obj.O = { value: 'y', enumerable: true }, obj[v.zlama] = { value: 'z', enumerable: true }, obj.I = { value: 'z', enumerable: true }, obj[v.hbasaEsasa] = { value: '{', enumerable: true }, obj[v.rwaha] = { value: '|', enumerable: true }, obj[d.qushaya] = { value: '', enumerable: true }, obj[d.rukkakha] = { value: ',', enumerable: true }, obj[d.lineaOccultans] = { value: '', enumerable: true }, obj[d.seyame] = { value: '', enumerable: true }, obj['`'] = { value: '', enumerable: true }, obj['1'] = { value: '', enumerable: true }, obj['8'] = { value: '', enumerable: true }, obj['#'] = { value: ',', enumerable: true }, obj['%'] = { value: '', enumerable: true }, obj['^'] = { value: '', enumerable: true }, obj['?'] = { value: '', enumerable: true }, obj['.'] = { value: ',', enumerable: true }, obj._ = { value: '', enumerable: true }, obj['-'] = { value: '', enumerable: true }, obj))
);

/**
 * Vowels and diacritics: used for consonantal only mapping
 * @constant
 * @type { Array.<string> }
 */
var dotting = Object.freeze(allVowels.concat(allDiacritics));

/**
 * Is character c an Estrangela consonant
 * @param { string } c input character
 * @returns { boolean } true if c is Estrangela consonant
 */
var isConsonant = function (c) { return allConsonants.indexOf(c) > -1; };

/**
 * Is character c an Estrangela vowel
 * @param { string } c input character
 * @returns { boolean } true if c is Estrangela vowel
 */
var isVowel = function (c) { return allVowels.indexOf(c) > -1; };

/**
 * Is character c an Estrangela diacritic
 * @param { string } c input character
 * @returns { boolean } true if c is Estrangela diacritic
 */
var isDiacritic = function (c) { return allDiacritics.indexOf(c) > -1; };

/**
 * Is character c an Estrangela punctuation
 * @param { string } c input character
 * @returns { boolean } true if c is Estrangela punctuation
 */
var isPunctuation = function (c) { return punctuation.indexOf(c) > -1; };

/**
 * Is character c an Estrangela joiner
 * @param { string } c input character
 * @returns { boolean } true if c is Estrangela joiner
 */
var isJoiner = function (c) { return joiners.indexOf(c) > -1; };

/**
 * Is character c an Estrangela Eastern Diacritic
 * @param { string } c input character
 * @returns { boolean } true if c is Estrangela Eastern Diacritic
 */
var isEasternDiacritic = function (c) { return easternDiacritics.indexOf(c) > -1; };

/**
 * Is character c an unconnected on the left Estrangela consonant
 * @param { string } c input character
 * @returns { boolean } true if c is unconnected on the left consonant
 */
var isUnconnected = function (c) { return unconnected.indexOf(c) > -1; };

/**
 * Returns true if c is dotting character
 * @param { string } c input character
 * @returns { boolean } true if c is dotting
 */
var isDotting = function (c) { return dotting.indexOf(c) > -1; };

/**
 * Map between regular and final ligature
 * @private
 */
var ligatureMap = Object.freeze(
  Object.create(null, {
    l: { value: 'f' },
    t: { value: 'F' }
  })
);

/**
 * Map between regular and final connected
 * @private
 */
var connectedMap = Object.freeze(
  Object.create(null, {
    b: { value: 'B' },
    g: { value: 'G' },

    x: { value: 'X' },
    '=': { value: '+' },
    y: { value: 'Y' },

    k: { value: 'K' },
    l: { value: 'L' },
    m: { value: 'M' },
    n: { value: 'N' },

    s: { value: 'S' },
    9: { value: '(' },
    p: { value: 'P' },

    q: { value: 'Q' },
    4: { value: '$' }
  })
);

/**
 * Map between final connected and final non connected
 * @private
 */
var nonConnectedMap = Object.freeze(
  Object.create(null, {
    K: { value: '<' },
    N: { value: 'J' }
  })
);

/**
 * Prototype for object managing final consonant setting
 * @private
 */
var finalProto = Object.create(null, {
  isAleph: { value: function (c) { return c === '0'; } },
  ligature: { value: function (c) { return ligatureMap[c]; } },
  connected: { value: function (c) { return connectedMap[c]; } },
  nonConnected: { value: function (c) { return nonConnectedMap[c]; } }
});

/**
 * Fix end consonant for a word in base Estrangela font
 * @param { string } word input word
 * @returns { string } word with end consonant corrected
 */
var endify = function (word) {
  if (!word) {
    return word;
  }

  var stack = [];
  var si = -1; // final consonant index in the stack
  var wi = -1; // index in the word of the final consonant
  var alephState = false;
  var nonConnectedState = false;
  var final = Object.create(finalProto);
  var i = word.length;
  while (--i >= 0) {
    var c = word.charAt(i);
    if (isConsonant(c)) {
      if (nonConnectedState) {
        var nc = final.nonConnected(stack[si]);
        if (nc && !final.connected(c)) {
          stack[si] = nc;
        }
        break;
      } else if (alephState) {
        var lc = final.ligature(c);
        if (lc) {
          wi = i;
          stack.push(lc);
        }
        break;
      } else {
        var fc = final.connected(c);
        if (fc) {
          wi = i;
          if (i === 0) {
            stack.push(final.nonConnected(fc) || fc); // standalone
            break;
          } else {
            si = stack.length;
            stack.push(fc);
            nonConnectedState = !!final.nonConnected(fc);
            if (!nonConnectedState) {
              break;
            }
          }
        } else {
          alephState = final.isAleph(c);
          if (!alephState || i === 0) {
            break; // not a final customized consonant, no change, get out
          }
        }
      }
    } else if (!nonConnectedState) {
      stack.push(c);
    }
  }
  if (wi >= 0) {
    var changedWord = wi > 0 ? word.substring(0, wi) : '';
    for (var j = stack.length - 1; j >= 0; --j) {
      changedWord += stack[j];
    }
    return changedWord;
  }
  return word;
};

/**
 * Return true if input word has vowels or diacritics
 * @param { string } word input word
 * @returns { boolean } true if word has vowels or diacritics
 */
var isDotted = hasDotting(isDotting);

/**
 * Remove dotting (vowels and diacritics), leaving consonantal word only.
 * @param { string } word input word to be processed
 * @returns { string } consonantal word
 */
var removeDotting = clearDotting(isDotting);

/**
 * Comparator function to be used for sorting Estrangela words
 * @static
 * @param { string } word1 first word to compare
 * @param { string } word2 second word to compare
 * @returns { number } -1, 0, 1 depending on word sorting
 */
var sort = getSort(letterAsciiMap, isConsonant);

export { consonantsByName, baseConsonants, finalConnected, finalNonConnected, ligatures, allConsonants, unconnected, vowelsByName, commonVowels, easternVowels, shiftedVowels, baseVowels, allVowels, diacriticsByName, commonDiacritics, easternDiacritics, baseDiacritics, shiftedDiacritics, allDiacritics, punctuation, joiners, letterAsciiMap, dotting, isConsonant, isVowel, isDiacritic, isPunctuation, isJoiner, isEasternDiacritic, isUnconnected, isDotting, endify, isDotted, removeDotting, sort };
